'''devel try-outs'''

import binascii
import socket
from pprint import pprint  # TODO DEVEL
import pdb  # TODO DEVEL

import certifi
import pyasn1_modules.rfc5280
from hexdump import hexdump
from pyasn1.codec.der.encoder import encode as der_encoder
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.type.univ import OctetString
from OpenSSL import SSL, crypto

from utlz import flo

from ctutlz.sct import Sct
from ctutlz.sctlist import SignedCertificateTimestampList


def create_context():

    def verify_callback(conn, cert, errnum, depth, ok):
        '''
        Return:
          0 if okay  (like "zero error")
          != 0, else
        '''
        certsubject = crypto.X509Name(cert.get_subject())
        commonname = certsubject.commonName
        return bool(ok == 1)

    ctx = SSL.Context(SSL.SSLv23_METHOD)
    ctx.set_options(SSL.OP_NO_SSLv2)
    ctx.set_options(SSL.OP_NO_SSLv3)

    ctx.set_verify(SSL.VERIFY_PEER, verify_callback)
    ca_filename = certifi.where()
    ctx.load_verify_locations(ca_filename)

    ctx.ocsp_resps = []

    def ocsp_client_callback(connection, ocsp_data, data):
        ctx.ocsp_resps.append(ocsp_data)
        return True

    ctx.set_ocsp_client_callback(ocsp_client_callback, data=None)

    return ctx


def create_socket():
    ctx = create_context()
    raw_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    return SSL.Connection(ctx, raw_sock)


def do_handshake(domain):
    # TODO: context in blocking mode
    # TODO: get tls extension
    #       https://github.com/openssl/openssl/blob/master/apps/s_client.c#L1634
    sock = create_socket()

    sock.request_ocsp()

    ocsp_response = None

    try:
        sock.connect((domain, 443))
        res = sock.do_handshake()
        chain = sock.get_peer_cert_chain()
        certificate = sock.get_peer_certificate()
        protocol_version = sock.get_protocol_version()
        cipher = sock.get_cipher_name()

#        from pprint import pprint
#        pprint(dir(sock))  # TODO DEBUG
        ctx = sock.get_context()
#        pprint(ctx.ocsp_res)
        if ctx.ocsp_resps:
            ocsp_response = ctx.ocsp_resps[0]

        # open('ocsp_res', 'wb').write(ocsp_response)

    except Exception as exc:
        import traceback
        tb = traceback.format_exc()
        exc_type = type(exc)
        print(flo('### {exc}\n\n{tb}\n\nexc-type: {exc_type}'))
        raise exc
    finally:
        sock.close()  # sock.close() possible?

    return certificate, chain, ocsp_response


def devel():
    # if False:
    if True:
        cert_x509, chain_x509s, ocsp_resp_der = do_handshake('www.google.com')
        # Deutsche Bank (EV-Zertifikat)
        cert_x509, chain_x509s, ocsp_resp_der = do_handshake('www.db.com')

        # print(chain_x509s) list of x509 entries

        # pprint(crypto.dump_certificate(type=crypto.FILETYPE_PEM, cert=cert_x509))  # as PEM
        # pprint(crypto.dump_certificate(type=crypto.FILETYPE_ASN1, cert=cert_x509)) # as DER
        # print(crypto.dump_certificate(type=crypto.FILETYPE_TEXT, cert=cert_x509).decode('ascii'))  # human readable

        cert_der = crypto.dump_certificate(type=crypto.FILETYPE_ASN1,
                                           cert=cert_x509)
        # write cert to file
        with open('www.db.com.crt', 'wb') as fh:
            fh.write(cert_der)

        #with open('www.db.com.chain', 'wb') as fh:
        #    for cert in chain_x509s:
        #        fh.write(crypto.dump_certificate(type=crypto.FILETYPE_PEM,
        #                                         cert=cert))
        #for index, cert in enumerate(chain_x509s):
        #    with open(flo('www.db.com.chain_crt_{index}'), 'wb') as fh:
        #        fh.write(crypto.dump_certificate(type=crypto.FILETYPE_PEM,
        #                                         cert=cert))

        # return
    else:
        # read cert from file
        with open('www.db.com.crt', 'rb') as fhr:
            cert_der = fhr.read()

    scts = scts_from_cert(cert_der)
    show_scts(scts)


def cert_of_domain(domain):
    cert_x509, chain_x509s, ocsp_resp_der = do_handshake(domain)
    cert_der = crypto.dump_certificate(type=crypto.FILETYPE_ASN1,
                                       cert=cert_x509)
    # https://tools.ietf.org/html/rfc5246#section-7.4.2
    issuer_cert_x509 = chain_x509s[1]
    issuer_cert_der = crypto.dump_certificate(type=crypto.FILETYPE_ASN1,
                                              cert=issuer_cert_x509)
    return cert_der, issuer_cert_der, ocsp_resp_der


def scts_from_cert(cert_der):
    cert, _ = der_decoder(cert_der,
                          asn1Spec=pyasn1_modules.rfc5280.Certificate())
    scts = []

    # FIXME: use pyopenssl access method
    exts = [extension
            for extension
            in cert['tbsCertificate']['extensions']
            if str(extension['extnID']) == '1.3.6.1.4.1.11129.2.4.2']

    if len(exts) != 0:
        extension_sctlist = exts[0]

        os_inner_der = extension_sctlist['extnValue']  # type: OctetString()
        os_inner, _ = der_decoder(os_inner_der, OctetString())

        sctlist_hex = os_inner.prettyPrint().split('0x')[-1]  # FIXME: ugly way
        sctlist_der = binascii.unhexlify(sctlist_hex)
        sctlist = SignedCertificateTimestampList(sctlist_der)
        scts = [Sct(entry.sct_der) for entry in sctlist.sct_list]

    return scts


def show_scts(scts):
    for sct in scts:
        print(flo('log_id (PEM): {sct.log_id_pem}'))
        print(flo('version: {sct.version}'))
        print(flo('timestamp: {sct.timestamp}'))
        print(flo('extensions length: {sct.extensions_len_hex}'))
        print(flo('signature alg hash: {sct.signature_alg_hash_hex}'))
        print(flo('signature alg sign: {sct.signature_alg_sign_hex}'))

    print('\n---- known logs (accepted by chrome)\n')
    from ctutlz.log import get_log_list
    logs = get_log_list()
    for log in logs:
        print(flo('description: {log.description}'))
        print(flo('log_id (PEM): {log.id_pem}\n'))


if __name__ == '__main__':
    devel()



import base64
import logging
import os
import sys
from collections import namedtuple

from utlz import flo

from ctutlz.sct import Sct
from ctutlz.sct.ee_cert import EndEntityCert
from ctutlz.sct.signature_input import create_signature_input
from ctutlz.sctlist import TlsExtension18
from ctutlz.utils import run_cmd


Extension18Result = namedtuple(
    typename='Extension18Result',
    field_names=[
        'ee_cert',
        'extension_18',
        'hostname',
        'output',
        'exitcode',
        'func_name',
        'timeout',
        'max_try',
        'num_try',
    ]
)


def tls_output_2_eecert_and_extension_18(tls_handshake_output):
    '''Extract PEM encoded end entity certificate and PEM encoded server
    extension 18 (SCT) from `tls_handshake_output` and return them as a tuple.
    '''

    '''lines of the pem encoded ee-cert'''
    ee_cert = []  # end entity certificate
    ee_cert_append = False

    '''lines of the pem encoded server extension 18 (SCT)'''
    extension_18 = []
    extension_18_append = False

    for line in tls_handshake_output.split('\n'):
        if line == '-----BEGIN CERTIFICATE-----':
            ee_cert_append = True
        elif line == '-----END CERTIFICATE-----':
            ee_cert_append = False
        elif line == '-----BEGIN SERVERINFO FOR EXTENSION 18-----':
            extension_18_append = True
        elif line == '-----END SERVERINFO FOR EXTENSION 18-----':
            extension_18_append = False
        elif ee_cert_append:
            ee_cert.append(line)
        elif extension_18_append:
            extension_18.append(line)

    ee_cert_der = base64.b64decode('\n'.join(ee_cert))
    extension_18_der = base64.b64decode('\n'.join(extension_18))

    return ee_cert_der, extension_18_der


def scrape_tls_extension_18_CLUNKY(hostname, timeout=30, max_try=3):
    '''Try to download the serverinfo extension 18 by doing a TLS handshake and
    return the extracted SCT (decoded; in DER format) or None.
    '''
    func_name = sys._getframe().f_code.co_name  # name of this function
    openssl = os.environ.get('OPENSSL_CMD', 'openssl')
    cmd = flo('{openssl}  s_client  -serverinfo 18  '
              '-connect {hostname}:443  -servername {hostname}')
    lgr = logging.getLogger('ctutlz')
    lgr.debug(flo('run_cmd: {cmd}'))
    res = run_cmd(cmd, timeout, max_try)
    extsn_18 = None
    ee_cert = None
    if res.exitcode == 0:
        ee_cert, extsn_18 = tls_output_2_eecert_and_extension_18(res.stdout_str)
    return Extension18Result(ee_cert, extsn_18, hostname,
                             res.stdout_str + res.stderr_str, res.exitcode,
                             func_name, timeout, max_try, num_try=1)


def scrape_tls_extension_18(hostname, timeout=30, max_try=3):
    # FIXME: use pyopenssl instead
    return scrape_tls_extension_18_CLUNKY(**locals())


def scts_by_tls(hostname, timeout=30, max_try=3):
    scts_by_tls.sign_input_func = create_signature_input
    res = scrape_tls_extension_18(hostname, timeout, max_try)
    assert res.extension_18 is not None
    if res.extension_18:
        tls_extension_18 = TlsExtension18(res.extension_18)
        sct_list = tls_extension_18.sct_list
        scts = [Sct(entry.sct_der) for entry in sct_list]
        return EndEntityCert(res.ee_cert), scts
    return EndEntityCert(res.ee_cert), None



# -*- coding: utf-8 -*-
import struct

from utlz import flo, namedtuple, StructContext
from ctutlz.utils import encode_to_pem, to_hex


'''Access data of an SCT in DER format by field names.'''
_Sct = namedtuple(
    typename='Sct',
    field_names=[
        'der',  # "raw" SCT data structure; type: bytes

        # "raw" SCT fields; each of type: bytes
        'version',
        'log_id',
        'timestamp',
        'extensions_len',
        'extensions',
        'signature_alg_hash',
        'signature_alg_sign',
        'signature_len',
        'signature',
    ],
    lazy_vals={
        'log_id_pem': lambda self: encode_to_pem(self.log_id),  # type: str
        'version_hex': lambda self: to_hex(self.version),
        'timestamp_hex': lambda self: to_hex(self.timestamp),
        'extensions_len_hex': lambda self: to_hex(self.extensions_len),
        'signature_alg_hash_hex': lambda self: to_hex(self.signature_alg_hash),
        'signature_alg_sign_hex': lambda self: to_hex(self.signature_alg_sign),
    }
)


def Sct(sct_der):
    # cf. https://tools.ietf.org/html/rfc6962#section-3.2
    with StructContext(sct_der) as struct:
        data_dict = {
            'der': sct_der,
            'version':            struct.read('!B'),
            'log_id':             struct.read('!32s'),
            'timestamp':          struct.read('!Q'),
            'extensions_len':     struct.read('!H'),
            'extensions': None,
            'signature_alg_hash': struct.read('!B'),
            'signature_alg_sign': struct.read('!B'),
        }
        signature_len = struct.read('!H')
        signature = struct.read(flo('!{signature_len}s'))
    return _Sct(signature_len=signature_len,
                signature=signature,
                **data_dict)



'''Validate SCTs using the openssl command line tool (which is "clunky").

A lot of the functionality originally comes and have been learned from the
script sct-verify.py written by Pier Carlo Chiodi:
https://github.com/pierky/sct-verify/blob/master/sct-verify.py (under GPL)

He also described the SCT validation steps very well in his blog:
https://blog.pierky.com/certificate-transparency-manually-verify-sct-with-openssl/
'''

import argparse
import binascii
import logging
import struct
import sys
from contextlib import contextmanager

from pyasn1.codec import ber
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.type.univ import Sequence, Any, ObjectIdentifier, OctetString
from pyasn1_modules import rfc2560
from utlz import flo

from ctutlz.log import get_log_list
from ctutlz.sct.ee_cert import EndEntityCert, IssuerCert
from ctutlz.sctlist import SignedCertificateTimestampList
from ctutlz.sctlist_scrape_tls import scts_by_tls
from ctutlz.sct import Sct
from ctutlz.sct.validation import validate_scts
from ctutlz.sct.signature_input import create_signature_input_precert
from ctutlz.sct.signature_input import create_signature_input
from ctutlz.utils import to_hex
from ctutlz import devel


def sctlist_hex_from_ocsp_pretty_print(ocsp_resp):
    sctlist_hex = None
    splitted = ocsp_resp.split('<no-name>=1.3.6.1.4.1.11129.2.4.5', 1)
    if len(splitted) > 1:
        _, after = splitted
        _, sctlist_hex_with_rest = after.split('<no-name>=0x', 1)
        sctlist_hex, _ = sctlist_hex_with_rest.split('\n', 1)
    return sctlist_hex


def scts_by_ocsp(hostname):
    scts_by_ocsp.sign_input_func = create_signature_input

    scts = []
    cert_der, issuer_cert_der, ocsp_resp_der = devel.cert_of_domain(hostname)
    if ocsp_resp_der:
        ocsp_resp, _ = der_decoder(ocsp_resp_der, rfc2560.OCSPResponse())

        responseBytes = ocsp_resp.getComponentByName('responseBytes')
        response_os = responseBytes.getComponentByName('response')

        der_decoder.defaultErrorState = ber.decoder.stDumpRawValue
        response, _ = der_decoder(response_os, Sequence())

        sctlist_os_hex = sctlist_hex_from_ocsp_pretty_print(response.prettyPrint())

        if sctlist_os_hex:
            sctlist_os_der = binascii.unhexlify(sctlist_os_hex)
            # open('sctlist_by_ocsp', 'wb').write(sctlist_os_der)
            sctlist_os, _ = der_decoder(sctlist_os_der, OctetString())

            sctlist_hex = sctlist_os.prettyPrint().split('0x')[-1]  # FIXME: ugly way
            sctlist_der = binascii.unhexlify(sctlist_hex)
            sctlist = SignedCertificateTimestampList(sctlist_der)
            scts = [Sct(entry.sct_der) for entry in sctlist.sct_list]

    return EndEntityCert(cert_der, IssuerCert(issuer_cert_der)), scts


def scts_by_cert(hostname):
    scts_by_cert.sign_input_func = create_signature_input_precert
    from ctutlz import devel
    cert_der, issuer_cert_der, ocsp_resp_der = devel.cert_of_domain(hostname)
    scts = devel.scts_from_cert(cert_der)
    return EndEntityCert(cert_der,
                         issuer_cert=IssuerCert(issuer_cert_der)), scts


def show_signature(sct):
    lgr = logging.getLogger('ctutlz')
    sig_offset = 0
    while sig_offset < len(sct.signature):
        if len(sct.signature) - sig_offset > 16:
            bytes_to_read = 16
        else:
            bytes_to_read = len(sct.signature) - sig_offset
        sig_bytes = struct.unpack_from(flo('!{bytes_to_read}s'),
                                       sct.signature,
                                       sig_offset)[0]
        if sig_offset == 0:
            lgr.info('Signature : %s' % to_hex(sig_bytes))
        else:
            lgr.info('            %s' % to_hex(sig_bytes))
        sig_offset = sig_offset + bytes_to_read


def show_validation(vdn):
    sct = vdn.sct

    lgr = logging.getLogger('ctutlz')
    sct_log_id1, sct_log_id2 = [to_hex(val)
                                for val
                                in struct.unpack("!16s16s", sct.log_id)]
    lgr.info('=' * 59)
    lgr.info(flo('Version   : {sct.version_hex}'))
    lgr.info(flo('LogID     : {sct_log_id1}'))
    lgr.info(flo('            {sct_log_id2}'))
    with loglevel(logging.INFO):
        lgr.info(flo('LogID b64 : {sct.log_id_pem}'))
    lgr.info(flo('Timestamp : {sct.timestamp} ({sct.timestamp_hex})'))
    lgr.info(flo('Extensions: {sct.extensions_len} ({sct.extensions_len_hex})'))
    lgr.info(flo('Algorithms: {sct.signature_alg_hash_hex}/'
                 '{sct.signature_alg_sign} (hash/sign)'))

    show_signature(sct)

    log = vdn.log
    with loglevel(logging.INFO):
        if log is None:
            lgr.info('Log not found\n')
        else:
            lgr.info(flo('Log found : {log.description}'))
    if log is not None:
        lgr.info('Operator  : ' + ', '.join(log.operated_by))

    with loglevel(logging.INFO):
        if vdn.output:
            lgr.info(flo('Result    : {vdn.output}'))

    # FIXME: show openssl return value on error
    if vdn.cmd_res is not None:
        lgr.debug(str(vdn.cmd_res._asdict().get('cmd', '')) + '\n')


def run_actions(hostname, actions):
    lgr = logging.getLogger('ctutlz')
    with loglevel(logging.INFO):
        lgr.info(flo('# {hostname}\n'))

    logs = get_log_list()  # FIXME make as argument
    # TODO DEBUG
    for log in logs:
        from ctutlz.utils import digest_from_pem
        assert log.id_der == digest_from_pem(log.key)

    for scrape_scts in actions:
        with loglevel(logging.INFO):
            lgr.info(flo('## {scrape_scts.__name__}\n'))
        issuer_cert = None
        ee_cert, scts = scrape_scts(hostname)
        if ee_cert:
            lgr.debug('got certificate\n')

            # FIXME: kinda hacky, un-pythonic
            # IssuerCert or None
            issuer_cert = ee_cert.issuer_cert

        validations = validate_scts(ee_cert, scts, logs, issuer_cert,
                                    sign_input_func=scrape_scts.sign_input_func)
        if validations:
            for vdn in validations:
                show_validation(vdn)
        elif ee_cert is not None:
            lgr.info('no SCTs\n')


@contextmanager
def loglevel(level):
    logger = logging.getLogger('ctutlz')
    levels = {}
    for handler in logger.handlers:
        levels[handler] = handler.level
        handler.setLevel(level)
    yield
    for handler in logger.handlers:
        handler.setLevel(levels[handler])


def setup_logging(loglevel):
    logger = logging.getLogger('ctutlz')
    logger.setLevel(logging.DEBUG)
    try:
        # python 2.6
        handler = logging.StreamHandler(stream=sys.stdout)
    except TypeError:
        # since python 2.7
        handler = logging.StreamHandler()
    handler.setLevel(loglevel)
    logger.addHandler(handler)
    return logger


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('hostname', nargs='+',
                        help="host name of the server (example: 'ritter.vg')")

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument('--short',
                     dest='loglevel',
                     action='store_const',
                     const=logging.WARNING,
                     default=logging.INFO,  # default loglevel if nothing set
                     help='show short result and warnings/errors only')
    meg.add_argument('--debug',
                     dest='loglevel',
                     action='store_const',
                     const=logging.DEBUG,
                     help='show more for diagnostic purposes')

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument('--cert-only',
                     dest='actions',
                     action='store_const',
                     const=[scts_by_cert],
                     default=[scts_by_cert, scts_by_tls, scts_by_ocsp],
                     help='only validate SCTs included in the certificate')
    meg.add_argument('--tls-only',
                     dest='actions',
                     action='store_const',
                     const=[scts_by_tls],
                     help='only validate SCTs gathered from TLS handshake')
    meg.add_argument('--ocsp-only',
                     dest='actions',
                     action='store_const',
                     const=[scts_by_ocsp],
                     help='only validate SCTs gathered via OCSP request')
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    logger = setup_logging(args.loglevel)
    logger.debug(args)

    for host in args.hostname:
        run_actions(host, actions=args.actions)


if __name__ == '__main__':
    main()



'''devel try-outs


'''

import sys

import ssl
from datetime import datetime
import pytz
import OpenSSL
import socket


from ndg.httpsclient.subj_alt_name import SubjectAltName
from pyasn1.codec.der import decoder as der_decoder
import pyasn1


def get_subj_alt_name(peer_cert):
    '''
    Copied from ndg.httpsclient.ssl_peer_verification.ServerSSLCertVerification
    Extract subjectAltName DNS name settings from certificate extensions
    @param peer_cert: peer certificate in SSL connection.  subjectAltName
    settings if any will be extracted from this
    @type peer_cert: OpenSSL.crypto.X509
    '''
    # Search through extensions
    dns_name = []
    general_names = SubjectAltName()
    for i in range(peer_cert.get_extension_count()):
        ext = peer_cert.get_extension(i)
        ext_name = ext.get_short_name()
        if ext_name == "subjectAltName":
            # PyOpenSSL returns extension data in ASN.1 encoded form
            ext_dat = ext.get_data()
            decoded_dat = der_decoder.decode(ext_dat, asn1Spec=general_names)

            for name in decoded_dat:
                if isinstance(name, SubjectAltName):
                    for entry in range(len(name)):
                        component = name.getComponentByPosition(entry)
                        dns_name.append(str(component.getComponent()))
    return dns_name

def main():
    if len(sys.argv) < 2:
        print("Usage: %s hostname1 [hostname2] [hostname3] ..." % sys.argv[0])
        print('')
        print("Preparation:")
        print(" $ virtualenv venv")
        print(" $ . venv/bin/activate")
        print(" $ pip install pytz pyasn1 pyOpenSSL ndg-httpsclient")

    color = {
        False: "\033[31;1m",
        True: "\033[32;1m",
        'end': "\033[0m",
        'error': "\033[33;1m",
    }

    for server in sys.argv[1:]:
        server = server.encode('utf-8')
        ctx = OpenSSL.SSL.Context(ssl.PROTOCOL_TLSv1)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        x509 = None
        try:
            s.connect((server, 443))
            cnx = OpenSSL.SSL.Connection(ctx, s)
            cnx.set_tlsext_host_name(server)
            cnx.set_connect_state()
            cnx.do_handshake()

            x509 = cnx.get_peer_certificate()
            s.close()
        except Exception as e:
            print("%30s: %s%s%s" % (server, color['error'], e, color['end']))
            continue

        issuer = x509.get_issuer()
        issuer_corp = x509.get_issuer().organizationName
        issuer_url = x509.get_issuer().organizationalUnitName
        issuer_x509 = x509.get_issuer().commonName

        server_name = x509.get_subject().commonName
        server_name_ok = server_name == server


        from pprint import pprint
        print('##### ######')
        print(type(x509))
        pprint(dir(x509))
        #print([x509.get_extension(i).get_short_name() for i in range(x509.get_extension_count())])
        ct_precert_scts = [
            x509.get_extension(i)
            for i
            in range(x509.get_extension_count())
            if x509.get_extension(i).get_short_name() == b'ct_precert_scts'
        ][0]
        print('---')
        print(type(ct_precert_scts))
        pprint(dir(ct_precert_scts))
        pprint(dir(ct_precert_scts.get_data()))
        data_hex = ct_precert_scts.get_data()
        print(data_hex)
        print(len(data_hex))

        import binascii
        # data_der = binascii.unhexlify(data_hex)
        data_der = data_hex

        from ctutlz.sctlist import TlsExtension18
        l = TlsExtension18(data_der)

        import pdb
        pdb.set_trace()


        return  # TODO DEVEL

        try:
            subjectAltNames = get_subj_alt_name(x509)
        except pyasn1.error.PyAsn1Error:
            print('pyasn1.error.PyAsn1Error')
            subjectAltNames = []
        server_name_alt_ok = server in subjectAltNames
        if server_name_alt_ok:
            server_name_alt = server
        elif len(subjectAltNames) == 0:
            server_name_alt = None
        else:
            server_name_alt = subjectAltNames[0]

        if len(subjectAltNames) > 1:
            server_name_alt += " (+%i)" % (len(subjectAltNames) - 1)

        now = datetime.now(pytz.utc)
        begin = datetime.strptime(x509.get_notBefore().decode('ascii'), "%Y%m%d%H%M%SZ").replace(tzinfo=pytz.UTC)
        begin_ok = begin < now
        end = datetime.strptime(x509.get_notAfter().decode('ascii'), "%Y%m%d%H%M%SZ").replace(tzinfo=pytz.UTC)
        end_ok = end > now

        print("%30s: %s%30s%s (alt: %s%30s%s) begin=%s%s%s end=%s%s%s issuer=%s" % (server,
            color[server_name_ok], server_name, color['end'],
            color[server_name_alt_ok], server_name_alt, color['end'],
            color[begin_ok], begin.strftime("%d.%m.%Y"), color['end'],
            color[end_ok], end.strftime("%d.%m.%Y"), color['end'],
            issuer_corp)
        )


if __name__ == '__main__':
    main()



import json
import struct
try:
    # since python 3.6
    import urllib.request as urllib_request
except ImportError:
    import urllib as urllib_request
from os.path import join, isfile, dirname

from utlz import namedtuple, text_with_newlines, load_json

from ctutlz.utils import decode_from_pem, digest_from_pem, sha256_digest
from ctutlz.utils import encode_to_pem


Log = namedtuple(
    typename='Log',
    field_names=[  # each of type: str
        'description',
        'key',     # PEM encoded, type: str
        'url',
        'maximum_merge_delay',
        'operated_by',
    ],
    lazy_vals={
        'key_der': lambda self: decode_from_pem(self.key),  # type: bytes
        'id_der': lambda self: digest_from_pem(self.key),   # type: bytes
        'id_pem': lambda self: encode_to_pem(self.id_der),  # type: str
        'pubkey': lambda self: '\n'.join([                  # type: str
                              '-----BEGIN PUBLIC KEY-----',
                              text_with_newlines(text=self.key, line_length=64),
                              '-----END PUBLIC KEY-----'
        ]),
        'pubkey_hash': lambda self: sha256_digest(self.key_der),  # TODO DEVEL
    }
)


def Logs(log_dicts):
    return [Log(**kwargs) for kwargs in log_dicts]


def logs_with_operator_names(logs_dict):
    for log in logs_dict['logs']:
        operator_ids = log['operated_by']
        operator_names = [operator['name']
                          for operator
                          in logs_dict['operators']
                          if operator['id'] in operator_ids]
        log['operated_by'] = operator_names
    return Logs(logs_dict['logs'])


def download_log_list_accepted_by_chrome():
    '''Download json file with known logs accepted by chrome and return the
    logs as a list of `Log` items.
    '''
    url = 'https://www.certificate-transparency.org/known-logs/log_list.json'
    response = urllib_request.urlopen(url)
    response_str = response.read()
    try:
        data = json.loads(response_str)
    except TypeError:
        # python 3.x < 3.6
        data = json.loads(response_str.decode('utf-8'))
    return logs_with_operator_names(data)


def get_log_list():
    try:
        return get_log_list.logs  # singleton function attribute
    except AttributeError:
        package_basedir = dirname(dirname(__file__))
        filename = join(package_basedir, 'log_list.json')
        if isfile(filename):
            data = load_json(filename)
            get_log_list.logs = logs_with_operator_names(data)
        else:
            get_log_list.logs = download_log_list_accepted_by_chrome()
    return get_log_list.logs



import collections
import os
from tempfile import NamedTemporaryFile as tempfile

from utlz import flo

from ctutlz.utils import run_cmd


SctValidationResult = collections.namedtuple(
    typename='SctValidationResult',
    field_names=[
        'ee_cert',   # DER format
        'sct',       # type: Sct
        'log',       # type: Log
        'verified',  # True or False
        'output',    # type: str
        'cmd_res',   # CmdResult (for debugging) # FIXME remove
    ]
)


def find_log(sct, logs):
    for log in logs:
        if log.id_der == sct.log_id:
            return log
    return None


def verify_signature_CLUNKY(signature_input, signature, pubkey):
    # with-cascade requiered for python2.6 support
    with tempfile() as signature_input_file:
        with tempfile() as signature_file:
            with tempfile() as pubkey_file:

                signature_input_file.write(signature_input)
                signature_input_file.seek(0)

                signature_file.write(signature)
                signature_file.seek(0)

                pubkey_file.write(pubkey)
                pubkey_file.seek(0)

                openssl = os.environ.get('OPENSSL_CMD', 'openssl')
                cmd = flo('{openssl}  dgst -sha256 -verify {pubkey_file.name} '
                          '-signature {signature_file.name} '
                          '{signature_input_file.name}')
                res = run_cmd(cmd, timeout=30, max_try=3)
                # print(output)  # TODO DEBUG
                output = res.stdout_str + res.stderr_str
                if res.exitcode == 0:
                    return True, output, res
                return False, output, res


verify_signature = verify_signature_CLUNKY  # FIXME: use pyopenssl instead


def validate_sct(ee_cert, sct, logs, issuer_cert, sign_input_func):
    log = find_log(sct, logs)
    if log:
        verified, output, cmd_res = verify_signature(
            signature_input=sign_input_func(ee_cert, sct, issuer_cert),
            signature=sct.signature,
            pubkey=log.pubkey.encode('ascii')
        )
        return SctValidationResult(ee_cert, sct, log,
                                   verified, output, cmd_res)
    return SctValidationResult(ee_cert, sct, log,
                               verified=False, output='', cmd_res=None)


def validate_scts(ee_cert, scts, logs, issuer_cert, sign_input_func):
    if scts:
        return [validate_sct(ee_cert, sct, logs, issuer_cert, sign_input_func)
                for sct
                in scts]
    return []
